#!/usr/bin/python
# -*- coding: utf-8 -*-

import dbus
import dbus.service
import dbus.mainloop.glib
import os
import time, datetime
import threading
from apt_handler import AppActions, AptHandler, WorkitemError
from utils.enums import INTERFACE, INTERFACES
from policykit_service import PolicyKitService
import aptsources.sourceslist

CRONPATH = "/var/spool/anacron/"
SOURCESPATH = "/etc/apt/sources.list.d"


class WorkItem:
     def __init__(self, pkgname, action, kwargs):
        self.pkgname = pkgname
        self.action = action
        self.kwargs = kwargs


class WorkThread(threading.Thread):

    def __init__(self, dbusDaemon):
        threading.Thread.__init__(self)
        self.dbusDaemon = dbusDaemon
        print "KUM DBUS START ", type(self.dbusDaemon)

    def run(self):
        while(True):

            if len(self.dbusDaemon.worklist) == 0:
                time.sleep(0.2)
                continue

            self.dbusDaemon.mutex.acquire()
            item = self.dbusDaemon.worklist.pop(0)
            self.dbusDaemon.mutex.release()
            try:
                func = getattr(self.dbusDaemon.aptHandler, item.action)
                if func is None:
                    print "Error action: ", item
                res = func(item.pkgname,item.kwargs)
                if res is False:
                    print "Action exec failed..."
            except WorkitemError as e:
                # 如果是包列表，此时强制装换为str，后面显示str将出现异常，故而暂不显示包列表
                kwarg = {"apt_appname": str(item.pkgname),
                        "apt_percent": str(-e.errornum),
                        "action": str(item.action),
                        }
                self.dbusDaemon.kum_apt_signal("apt_error", kwarg)
            except:
                kwarg = {"apt_appname": str(item.pkgname),
                        "apt_percent": '-24',
                        "action": str(item.action),
                        }
                self.dbusDaemon.kum_apt_signal("apt_error", kwarg)
            time.sleep(0.3)


class Daemon(PolicyKitService):

    def __init__ (self, bus, mainloop):
        self.aptHandler = AptHandler(self)
        self.worklist = []
        self.cancel_name_list = []
        self.mutex = threading.RLock()
        self.cancelmutex = threading.RLock()
        self.worker_thread = WorkThread(self)
        self.worker_thread.setDaemon(True)
        self.worker_thread.start()

        bus_name = dbus.service.BusName(INTERFACE, bus=bus)
        PolicyKitService.__init__(self, bus_name, INTERFACES)
        self.mainloop = mainloop

    def stop(self):
        """ stop the dbus controller and remove from the bus """
        self.remove_from_connection()

    def add_worker_item(self, item):
        print "####add_worker_item:", item
        self.mutex.acquire()
        self.worklist.append(item)
        self.mutex.release()

    def del_worker_item_by_name(self, pkgname):
        print "####del_worker_item_by_name:", pkgname
        exist = False
        self.mutex.acquire()
        for item in self.worklist:
            if item.pkgname == pkgname:
                exist = True
                break
        if exist is True:
            self.worklist.remove(pkgname)
        self.mutex.release()

        self.cancelmutex.acquire()
        self.cancel_name_list.append(pkgname)
        self.cancelmutex.release()
        print "####del_worker_item_by_name finished!"

    def check_cancel_worker_item(self, pkgname):
#        print "####check_cancel_worker_item:",pkgname
        cancel = False
        self.cancelmutex.acquire()
#        print "check_cancel_worker_item:",len(self.cancel_name_list)
        for item in self.cancel_name_list:
            if item == pkgname:
                cancel = True
                break
        if cancel is True:
            self.cancel_name_list.remove(pkgname)
        self.cancelmutex.release()

#        print "####check_cancel_worker_item finished!:",cancel
        return cancel

    @dbus.service.method(INTERFACE, in_signature='', out_signature='')
    def exit(self):
        self.mainloop.quit()

    # install package sa:kum_fetch_signal() and kum_apt_signal()
    @dbus.service.method(INTERFACE, in_signature='as', out_signature='b', sender_keyword='sender')
    def install_and_upgrade(self, pkgNames, sender=None):
        print "####install_and_upgrade: ",pkgNames

        item = WorkItem(pkgNames, AppActions.INS_AND_UP, None)

        self.add_worker_item(item)

        print "####install_and_upgrade return"
        return True

    # install package sa:kum_fetch_signal() and kum_apt_signal()
    @dbus.service.method(INTERFACE, in_signature='as', out_signature='b', sender_keyword='sender')
    def install(self, pkgNames, sender=None):
        print "####install: ",pkgNames

        item = WorkItem(pkgNames, AppActions.INSTALL, None)

        self.add_worker_item(item)

        print "####install return"
        return True

    @dbus.service.method(INTERFACE, in_signature='s', out_signature='b', sender_keyword='sender')
    def insone(self, pkgName, sender=None):
        print "####insone: ", pkgName

        item = WorkItem(pkgName, AppActions.INSTALL_ONE, None)

        self.add_worker_item(item)

        print "####insone return"
        return True

    # update package sa:kum_fetch_signal() and kum_apt_signal()
    @dbus.service.method(INTERFACE, in_signature='as', out_signature='b', sender_keyword='sender')
    def upgrade(self, pkgNames, sender=None):
        print "####upgrade: ", pkgNames

        item = WorkItem(pkgNames, AppActions.UPGRADE, None)

        self.add_worker_item(item)

        print "####upgrade return"
        return True

    @dbus.service.method(INTERFACE, in_signature='s', out_signature='b', sender_keyword='sender')
    def upone(self, pkgName, sender=None):
        print "####upone: ", pkgName

        item = WorkItem(pkgName, AppActions.UPGRADE_ONE, None)

        self.add_worker_item(item)

        print "####upone return"
        return True

    @dbus.service.method(INTERFACE, in_signature='s', out_signature='b', sender_keyword='sender')
    def cancel(self, pkgName, sender=None):
        print "####cancel: ",pkgName

        self.del_worker_item_by_name(pkgName)

        print "####cancel return"
        return True

    # apt-get update sa:kum_fetch_signal()
    @dbus.service.method(INTERFACE, in_signature='b', out_signature='b', sender_keyword='sender')
    def update(self, quiet, sender=None):
        print "####update: "

        kwargs = {"quiet":str(quiet),
                  }

        item = WorkItem("#update", AppActions.UPDATE, kwargs)

        self.add_worker_item(item)

        print "####update return"
        return True

    @dbus.service.method(INTERFACE, in_signature='b', out_signature='b', sender_keyword='sender')
    def update_by_shell(self, quiet, sender=None):
        print "####update by shell: "

        kwargs = {"quiet":str(quiet),
                  }

        item = WorkItem("#update-shell", AppActions.UPDATESHELL, kwargs)

        self.add_worker_item(item)

        print "####update by shell return"
        return True

    @dbus.service.method(INTERFACE, in_signature='b', out_signature='b', sender_keyword='sender')
    def configure_dpkg_by_shell(self, quiet, sender=None):
        print "####configure dpkg by shell: "

        kwargs = {"quiet":str(quiet),
                  }

        item = WorkItem("#configure-shell", AppActions.CONFIGURESHELL, kwargs)

        self.add_worker_item(item)

        print "####configure dpkg by shell return"
        return True

    #-------------------------------------signals----------------------------------------

    @dbus.service.signal(INTERFACE, signature='sa{ss}')
    def software_auth_signal(self, type, msg):
        pass

    # package download status signal
    '''parm mean
        type:
            start:start download
            stop:all work is finish
            done:all items download finished
            fail:download failed
            fetch:one item download finished
            pulse:download status, this msg given a string like dict
        msg:
            a message of type, sometimes is None
    '''
    @dbus.service.signal(INTERFACE, signature='sa{ss}')
    def kum_fetch_signal(self, type_msg, msg):
        pass

    # package install/update/remove signal
    '''parm mean
        type:
            start:start work
            stop:work finish
            error:got a error
            pulse:work status, this msg given a string like dict
        msg:
            a message of type, sometimes is None
    '''
    @dbus.service.signal(INTERFACE, signature='sa{ss}')
    def kum_apt_signal(self, type_msg, msg):
        pass

    # test func
    @dbus.service.method(INTERFACE, in_signature='i', out_signature='i')
    def test_return_value(self, arg):
        return 2

    @dbus.service.method(INTERFACE, in_signature='sssss', out_signature='b')
    def change_source_list_to_default(self, service_key, pwd, current_user, os_codename, available_source):
        if os.path.exists(SOURCESPATH):
            for item in os.listdir(SOURCESPATH):
                targetFile = os.path.join(SOURCESPATH, item)
                if os.path.isfile(targetFile):
                    os.remove(targetFile)

        cdrom_addr = "file:///media/" + current_user + "/Kylin-4.0"
        ginkgo_addr = "http://" + service_key + ":" + pwd + "@archive.kylinos.cn/"
        juniper_addr = "http://" + service_key + ":" + pwd + "@archive.kylinos.cn/kylin/juniper"

        if available_source == "no_source":
            fp = open('/etc/apt/sources.list', 'w+')
            fp.tell()
            if os_codename == "ginkgo":
                fp.write("deb " + cdrom_addr + " ginkgo main restricted multiverse universe\n")
                fp.write("deb " + ginkgo_addr + " ginkgo main restricted multiverse universe\n")
            if os_codename == "juniper":
                fp.write("deb " + cdrom_addr + " juniper main restricted multiverse universe\n")
                fp.write("deb " + juniper_addr + " juniper main restricted multiverse universe\n")
            fp.close()

        if available_source == "dvd_source":
            fp = open('/etc/apt/sources.list', 'a')
            if os_codename == "ginkgo":
                fp.write("deb " + ginkgo_addr + " ginkgo main restricted multiverse universe\n")
            if os_codename == "juniper":
                fp.write("deb " + juniper_addr + " juniper main restricted multiverse universe\n")
            fp.close()

        if available_source == "ftp_source":
            fp = open('/etc/apt/sources.list', 'a')
            if os_codename == "ginkgo":
                fp.write("deb " + cdrom_addr + " ginkgo main restricted multiverse universe\n")
                fp.write("deb " + ginkgo_addr + " ginkgo main restricted multiverse universe\n")
            if os_codename == "juniper":
                fp.write("deb " + cdrom_addr + " juniper main restricted multiverse universe\n")
                fp.write("deb " + juniper_addr + " juniper main restricted multiverse universe\n")
            fp.close()

        if available_source == "http_source":
            fp = open('/etc/apt/sources.list', 'a')
            if os_codename == "ginkgo":
                fp.write("deb " + cdrom_addr + " ginkgo main restricted multiverse universe\n")
            if os_codename == "juniper":
                fp.write("deb " + cdrom_addr + " juniper main restricted multiverse universe\n")
            fp.close()

        return True

    @dbus.service.method(INTERFACE, in_signature='as', out_signature='b')
    def change_source_list_to_kylin_update_server(self, lines):
        if os.path.exists(SOURCESPATH):
           for item in os.listdir(SOURCESPATH):
               targetFile = os.path.join(SOURCESPATH, item)
               if os.path.isfile(targetFile):
                   os.remove(targetFile)

        fp = open('/etc/apt/sources.list', 'w+')
        fp.tell()
        for line in lines:
           fp.write(line + "\n")
        fp.close()

        return True